// ======================================================================
//
// object.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Object.h"

#include "Appear.h"
#include "Camera.h"
#include "Control.h"
#include "Extents.h"
#include "Iff.h"
#include "Watcher.h"

// ======================================================================

int Object::nextUniqueId = 1;

const Tag TAG_OBJ = TAG3(O,B,J);

// ======================================================================

Object::Object(ObjectId newObjId, const char *appearName, Control *newControl, int newUniqueId)
: objectId(newObjId), objectToWorld()
{
	init(newUniqueId);

	if (appearName)
		setAppear(appearName);

	if (newControl)
		setControl(newControl);
}

// ----------------------------------------------------------------------

Object::~Object(void)
{
	// delete all children
	while (children)
	{
		// -q- child will detach itself from its parent
		delete children->obj;
	}

	// detach object from any existing parent
	detachFromParent();

	// release the appearance
	Appear::removeUser(appear);
	appear = NULL;

	delete control;
	control = NULL;

	delete [] name;
	name = NULL;

	if (watchable)
		BaseWatcher::kill(this);
}

// ----------------------------------------------------------------------

void Object::init(int newUniqueId)
{
	listId         = 0;
	appear         = NULL;
	control        = NULL;
	name           = NULL;
	parent         = NULL;
	children       = NULL;
	rotationCount  = 0;
	active         = true;
	watchable      = true;

	xform = &objectToWorld;
	objectToWorld.identify();

	if (!newUniqueId)
		uniqueId = nextUniqueId++;
	else
		uniqueId = newUniqueId;
}

// ----------------------------------------------------------------------

void Object::activate(void)
{
	active = true;
}

// ----------------------------------------------------------------------

void Object::deactivate(void)
{
	active = false;
}

// ----------------------------------------------------------------------

void Object::setControl(Control *newControl)
{
	if (control != newControl)
	{
		if (control)
			delete control;
		control = newControl;
		if (control)
			control->setOwner(this);
	}
}

// ----------------------------------------------------------------------

void Object::setName(const char *n)
{
	delete [] name;
	name = NULL;

	// set the new name
	if (n)
	{
		name = NEW char[strlen(n)+1];
		strcpy(name, n);
	}
}

// ----------------------------------------------------------------------

void Object::setAppear(Appear *newAppear)
{
	if (appear != newAppear)
	{
		Appear::removeUser(appear);
		appear = Appear::addUser(newAppear);
	}
}

// ----------------------------------------------------------------------

void Object::setAppear(const char *appearName, real scale)
{
	setAppear(Appear::find(appearName, scale, false));
}

// ----------------------------------------------------------------------

void Object::attachChild(Object *child, bool renderChild)
{
	ChildLink *childLink;

	// handle child already attached
	if (child->parent == this)
		return;

	// unattach the child from any previous parent it had
	if (child->parent)
		child->detachFromParent();

	// hook the child in to the children
	childLink = NEW ChildLink;
	childLink->obj = child;
	childLink->next = children;
	childLink->render = renderChild;
	children = childLink;

	// let the child know we are its parent
	child->parent = this;
	child->xform = NEW Xform(child->objectToWorld);
	child->recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::detachChild(Object *child, bool required)
{
	ChildLink *front, *back;

#ifdef NDEBUG
		// -q- keep required from being unreferenced in non-debug builds
		((void)required);
#endif

	// search for the childlink that contains this child
	for (back = NULL, front = children;
			 front && front->obj != child;
			 back = front, front = front->next)
		;

	if (!front)
	{
		DEBUG_FATAL(required, ("Object::detach child not found"));
		return;
	}

	// remove this childlink
	if (back)
		back->next = front->next;
	else
		children = front->next;

	// free the memory in the childlink
	delete front;

	// child has left its parent
	child->parent = NULL;

	// leave it in the world where it was
	delete child->xform;
	child->xform = &child->objectToWorld;
}

// ----------------------------------------------------------------------

void Object::detachFromParent(void)
{
	if (parent)
		parent->detachChild(this, 1);
}

// ----------------------------------------------------------------------

void Object::recalcWorldXform(void)
{
	ChildLink *childLink;

	// handle parental heirarchies
	if (parent)
		objectToWorld.multiply(parent->objectToWorld, *xform);

	// recursively handle any children
	for (childLink = children; childLink; childLink = childLink->next)
		childLink->obj->recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::prepare(real time)
{
	if (control)
		control->prepare(time);
}

// ----------------------------------------------------------------------

void Object::update(real time)
{
	Object *object;

	// recursively handle any children
	for (ChildLink *childLink = children; childLink;)
	{
		// get the object and advance to the next child object link
		object = childLink->obj;
		childLink = childLink->next;

		object->update(time);
	}
}

// ----------------------------------------------------------------------

int Object::inactiveAlter(real time)
{
	Object *object;

	// should we alter appearances on inactive objects? 
	//	maybe call an inactive alter for an appearance?
	//	if (appear)
	//		appear->alter(time);

	// recursively handle any children
	for (ChildLink *childLink = children; childLink;)
	{
		// get the object and advance to the next child object link
		object = childLink->obj;
		childLink = childLink->next;

		// inactivealter the object and delete it if it wants to die
		if (!object->inactiveAlter(time))
			delete object;
	}

	return 1;
}

// ----------------------------------------------------------------------

int Object::alter(real time)
{
	Object *object;

	if (appear)
		appear->alter(time);

	// recursively handle any children
	for (ChildLink *childLink = children; childLink;)
	{
		// get the object and advance to the next child object link
		object = childLink->obj;
		childLink = childLink->next;

		// alter the object and delete it if it wants to die
		if (!object->alter(time))
			delete object;
	}

	return 1;
}

// ----------------------------------------------------------------------

void Object::createPacket(Iff &iff, bool first)
{
	Xform::Save save;

	if (first)
		iff.appendChunk(TAG_OBJ, 0, NULL);

	xform->writeSave(&save);
	iff.appendToChunk(&save, sizeof(save));

	if (control)
		control->createPacket(iff);
}

// ----------------------------------------------------------------------

void Object::processPacket(Iff &iff, bool first)
{
	if (first)
		iff.seekChunk(TAG_OBJ);

	iff.read_floatXform(1, xform);

	if (first)
		iff.exitChunk();
}

// ----------------------------------------------------------------------

bool Object::render(bool)
{
	bool result;
	ChildLink *childLink;

	for (childLink = children, result = false; childLink; childLink = childLink->next)
		if (childLink->obj->render(true))
			result = true;

	return result;
}

// ----------------------------------------------------------------------

void Object::addToCamera(Camera &camera)
{
	if (!camera.isExcluded(this))
	{
		if (appear)
		{
			const Appear *addAppear;

			if ((addAppear = appear->getRenderAppear(camera, this)) != NULL)
				camera.addObject(this, addAppear);
		}

		// recursively handle any children
		for (ChildLink *childLink = children; childLink; childLink = childLink->next)
			if (childLink->render)
				childLink->obj->addToCamera(camera);
	}
}

// ----------------------------------------------------------------------

void Object::move(const Vector &vec)
{
	Vector v;

	xform->rotate(vec, &v);
	xform->moveRelative(v);
	recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::move(real dx, real dy, real dz)
{
	move(Vector(dx, dy, dz));
}

// ----------------------------------------------------------------------

void Object::rotateReset(bool childrenToo)
{
	xform->rotateReset();
	rotationCount = 0;
	recalcWorldXform();

	if (childrenToo)
	{
		ChildLink *childLink;

		for (childLink = children; childLink; childLink = childLink->next)
			childLink->obj->rotateReset(true);
	}
}

// ----------------------------------------------------------------------

void Object::rotateToFrame(const Vector *x, const Vector *y, const Vector *z)
{
	xform->rotateToFrame(x, y, z);
	recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::rotateAbsoluteYPR(real yawRadians, real pitchRadians, real rollRadians)
{
	xform->rotateReset();
	rotationCount = 0;
	rotateYPR(yawRadians, pitchRadians, rollRadians);
}

// ----------------------------------------------------------------------

void Object::rotateYPR(real yawRadians, real pitchRadians, real rollRadians)
{
	if (yawRadians != CONST_REAL(0))
	{
		xform->yaw(yawRadians);
		reorthogonalize();
	}

	if (pitchRadians != CONST_REAL(0))
	{
		xform->pitch(pitchRadians);
		reorthogonalize();
	}

	if (rollRadians != CONST_REAL(0))
	{
		xform->roll(rollRadians);
		reorthogonalize();
	}

	recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::yaw(real radians)
{
	xform->yaw(radians);
	reorthogonalize();
	recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::pitch(real radians)
{
	xform->pitch(radians);
	reorthogonalize();
	recalcWorldXform();
}

// ----------------------------------------------------------------------

void Object::roll(real radians)
{
	xform->roll(radians);
	reorthogonalize();
	recalcWorldXform();
}

// ----------------------------------------------------------------------

bool Object::detectCollision(Collision *col)
{
	bool       result;
	ChildLink *childLink;

	// recursively handle any children
	for (childLink = children; childLink; childLink = childLink->next)
		if ((result = childLink->obj->detectCollision(col)) != 0)
		 return result;

	return false;
}

// ----------------------------------------------------------------------

void Object::takeDamage(const Vector &, real)
{
}

// ----------------------------------------------------------------------

void Object::applyDamage(Object *, const Vector &)
{
}

// ----------------------------------------------------------------------

real Object::getSphereRadius(void) const
{
	if (appear && appear->getExtents())
		return appear->getExtents()->getSphereRadius();

	return CONST_REAL(-1);
}

// ----------------------------------------------------------------------

const Vector &Object::getSphereCenter(void) const
{
	if (appear && appear->getExtents())
		return appear->getExtents()->getSphereCenter();

	return Vector::zero;
}

// ----------------------------------------------------------------------

const Extents *Object::getExtents(void) const
{
	if (appear)
		return appear->getExtents();

	return NULL;
}

// ======================================================================
